@@ -20,7 +20,9 @@
 package devicestate_test
 
 import (
+	"bytes"
 	"errors"
+	"fmt"
 	"os"
 	"path/filepath"
 
@@ -28,7 +30,9 @@ import (
 
 	"github.com/snapcore/snapd/asserts"
 	"github.com/snapcore/snapd/asserts/assertstest"
+	"github.com/snapcore/snapd/boot"
 	"github.com/snapcore/snapd/dirs"
+	"github.com/snapcore/snapd/logger"
 	"github.com/snapcore/snapd/overlord/auth"
 	"github.com/snapcore/snapd/overlord/devicestate"
 	"github.com/snapcore/snapd/overlord/devicestate/devicestatetest"
@@ -49,6 +53,7 @@ type mockedSystemSeed struct {
 type deviceMgrSystemsSuite struct {
 	deviceMgrBaseSuite
 
+	logbuf            *bytes.Buffer
 	mockedSystemSeeds []mockedSystemSeed
 }
 
@@ -184,6 +189,20 @@ func (s *deviceMgrSystemsSuite) SetUpTest(c *C) {
 	// all tests should be in run mode by default, if they need to be in
 	// different modes they should set that individually
 	devicestate.SetSystemMode(s.mgr, "run")
+
+	// state after mark-seeded ran
+	modeenv := boot.Modeenv{
+		Mode:           "run",
+		RecoverySystem: "",
+	}
+	err := modeenv.WriteTo("")
+	s.state.Set("seeded", true)
+
+	c.Assert(err, IsNil)
+
+	logbuf, restore := logger.MockLogger()
+	s.logbuf = logbuf
+	s.AddCleanup(restore)
 }
 
 func (s *deviceMgrSystemsSuite) TestListNoSystems(c *C) {
@@ -311,7 +330,7 @@ func (s *deviceMgrSystemsSuite) TestBrokenSeedSystems(c *C) {
 
 func (s *deviceMgrSystemsSuite) TestRequestModeInstallHappyForAny(c *C) {
 	// no current system
-	err := s.mgr.RequestSystemAction("20191119", devicestate.SystemAction{Mode: "install"})
+	err := s.mgr.RequestSystemAction("20191119", devicestate.SystemAction{Mode: "install", Title: "Install"})
 	c.Assert(err, IsNil)
 
 	m, err := s.bootloader.GetBootVars("snapd_recovery_mode", "snapd_recovery_system")
@@ -321,6 +340,7 @@ func (s *deviceMgrSystemsSuite) TestRequestModeInstallHappyForAny(c *C) {
 		"snapd_recovery_mode":   "install",
 	})
 	c.Check(s.restartRequests, DeepEquals, []state.RestartType{state.RestartSystemNow})
+	c.Check(s.logbuf.String(), Matches, `.*: restarting into system "20191119" for action "Install"\n`)
 }
 
 func (s *deviceMgrSystemsSuite) TestRequestSameModeSameSystem(c *C) {
@@ -336,12 +356,25 @@ func (s *deviceMgrSystemsSuite) TestRequestSameModeSameSystem(c *C) {
 
 	label := s.mockedSystemSeeds[0].label
 
-	happyModes := []string{"run", "recover"}
-	sadModes := []string{"install"}
+	happyModes := []string{"run"}
+	sadModes := []string{"install", "recover"}
 
 	for _, mode := range append(happyModes, sadModes...) {
+		s.logbuf.Reset()
+
+		c.Logf("checking mode: %q", mode)
+		// non run modes use modeenv
+		modeenv := boot.Modeenv{
+			Mode: mode,
+		}
+		if mode != "run" {
+			modeenv.RecoverySystem = s.mockedSystemSeeds[0].label
+		}
+		err := modeenv.WriteTo("")
+		c.Assert(err, IsNil)
+
 		devicestate.SetSystemMode(s.mgr, mode)
-		err := s.bootloader.SetBootVars(map[string]string{
+		err = s.bootloader.SetBootVars(map[string]string{
 			"snapd_recovery_mode":   mode,
 			"snapd_recovery_system": label,
 		})
@@ -361,28 +394,110 @@ func (s *deviceMgrSystemsSuite) TestRequestSameModeSameSystem(c *C) {
 		})
 		// should never restart
 		c.Check(s.restartRequests, HasLen, 0)
+		// no log output
+		c.Check(s.logbuf.String(), Equals, "")
 	}
 }
 
-func (s *deviceMgrSystemsSuite) TestRequestNotYetSeeded(c *C) {
-	// seeded-systems is unset
+func (s *deviceMgrSystemsSuite) TestRequestSeedingSameConflict(c *C) {
 	label := s.mockedSystemSeeds[0].label
 
+	devicestate.SetSystemMode(s.mgr, "run")
+
+	s.state.Lock()
+	s.state.Set("seeded", nil)
+	s.state.Unlock()
+
 	for _, mode := range []string{"run", "install", "recover"} {
-		devicestate.SetSystemMode(s.mgr, mode)
-		err := s.bootloader.SetBootVars(map[string]string{
-			"snapd_recovery_mode":   mode,
+		s.logbuf.Reset()
+
+		c.Logf("checking mode: %q", mode)
+		modeenv := boot.Modeenv{
+			Mode:           mode,
+			RecoverySystem: s.mockedSystemSeeds[0].label,
+		}
+		err := modeenv.WriteTo("")
+		c.Assert(err, IsNil)
+
+		err = s.bootloader.SetBootVars(map[string]string{
+			"snapd_recovery_mode":   "",
 			"snapd_recovery_system": label,
 		})
 		c.Assert(err, IsNil)
 		err = s.mgr.RequestSystemAction(label, devicestate.SystemAction{Mode: mode})
-		c.Assert(err, Equals, devicestate.ErrUnsupportedAction)
+		c.Assert(err, ErrorMatches, "cannot request system action, system is seeding")
+		// no log output
+		c.Check(s.logbuf.String(), Equals, "")
+	}
+}
+
+func (s *deviceMgrSystemsSuite) TestRequestSeedingDifferentNoConflict(c *C) {
+	label := s.mockedSystemSeeds[0].label
+	otherLabel := s.mockedSystemSeeds[1].label
+
+	devicestate.SetSystemMode(s.mgr, "run")
+
+	modeenv := boot.Modeenv{
+		Mode:           "run",
+		RecoverySystem: label,
+	}
+	err := modeenv.WriteTo("")
+	c.Assert(err, IsNil)
+
+	s.state.Lock()
+	s.state.Set("seeded", nil)
+	s.state.Unlock()
+
+	// we can only go to install mode of other system when one is currently
+	// being seeded
+	err = s.bootloader.SetBootVars(map[string]string{
+		"snapd_recovery_mode":   "",
+		"snapd_recovery_system": label,
+	})
+	c.Assert(err, IsNil)
+	err = s.mgr.RequestSystemAction(otherLabel, devicestate.SystemAction{Mode: "install"})
+	c.Assert(err, IsNil)
+	m, err := s.bootloader.GetBootVars("snapd_recovery_mode", "snapd_recovery_system")
+	c.Assert(err, IsNil)
+	c.Check(m, DeepEquals, map[string]string{
+		"snapd_recovery_system": otherLabel,
+		"snapd_recovery_mode":   "install",
+	})
+	c.Check(s.logbuf.String(), Matches, fmt.Sprintf(`.*: restarting into system "%s" for action "Install"\n`, otherLabel))
+}
+
+func (s *deviceMgrSystemsSuite) testRequestModeWithRestart(c *C, toModes []string, label string) {
+	for _, mode := range toModes {
+		c.Logf("checking mode: %q", mode)
+		err := s.mgr.RequestSystemAction(label, devicestate.SystemAction{Mode: mode})
+		c.Assert(err, IsNil)
+		m, err := s.bootloader.GetBootVars("snapd_recovery_mode", "snapd_recovery_system")
+		c.Assert(err, IsNil)
+		c.Check(m, DeepEquals, map[string]string{
+			"snapd_recovery_system": label,
+			"snapd_recovery_mode":   mode,
+		})
+		c.Check(s.restartRequests, DeepEquals, []state.RestartType{state.RestartSystemNow})
+		s.restartRequests = nil
+		s.bootloader.BootVars = map[string]string{}
+
+		// TODO: also test correct action string logging
+		c.Check(s.logbuf.String(), Matches, fmt.Sprintf(`.*: restarting into system "%s" for action ".*"\n`, label))
+		s.logbuf.Reset()
 	}
 }
 
 func (s *deviceMgrSystemsSuite) TestRequestModeRunInstallForRecover(c *C) {
 	// we are in recover mode here
 	devicestate.SetSystemMode(s.mgr, "recover")
+	// non run modes use modeenv
+	modeenv := boot.Modeenv{
+		Mode:           "recover",
+		RecoverySystem: s.mockedSystemSeeds[0].label,
+	}
+	err := modeenv.WriteTo("")
+	c.Assert(err, IsNil)
+
 	s.state.Lock()
 	s.state.Set("seeded-systems", []devicestate.SeededSystem{
 		{
@@ -393,23 +508,18 @@ func (s *deviceMgrSystemsSuite) TestRequestModeRunInstallForRecover(c *C) {
 	})
 	s.state.Unlock()
 
-	for _, mode := range []string{"install", "run"} {
-		err := s.mgr.RequestSystemAction(s.mockedSystemSeeds[0].label,
-			devicestate.SystemAction{Mode: mode})
-		c.Assert(err, IsNil)
-		m, err := s.bootloader.GetBootVars("snapd_recovery_mode", "snapd_recovery_system")
-		c.Assert(err, IsNil)
-		c.Check(m, DeepEquals, map[string]string{
-			"snapd_recovery_system": s.mockedSystemSeeds[0].label,
-			"snapd_recovery_mode":   mode,
-		})
-		c.Check(s.restartRequests, DeepEquals, []state.RestartType{state.RestartSystemNow})
-		s.restartRequests = nil
-		s.bootloader.BootVars = map[string]string{}
-	}
+	s.testRequestModeWithRestart(c, []string{"install", "run"}, s.mockedSystemSeeds[0].label)
 }
 
 func (s *deviceMgrSystemsSuite) TestRequestModeInstallRecoverForCurrent(c *C) {
+	devicestate.SetSystemMode(s.mgr, "run")
+	// non run modes use modeenv
+	modeenv := boot.Modeenv{
+		Mode: "run",
+	}
+	err := modeenv.WriteTo("")
+	c.Assert(err, IsNil)
+
 	s.state.Lock()
 	s.state.Set("seeded-systems", []devicestate.SeededSystem{
 		{
@@ -420,20 +530,7 @@ func (s *deviceMgrSystemsSuite) TestRequestModeInstallRecoverForCurrent(c *C) {
 	})
 	s.state.Unlock()
 
-	for _, mode := range []string{"install", "recover"} {
-		err := s.mgr.RequestSystemAction(s.mockedSystemSeeds[0].label,
-			devicestate.SystemAction{Mode: mode})
-		c.Assert(err, IsNil)
-		m, err := s.bootloader.GetBootVars("snapd_recovery_mode", "snapd_recovery_system")
-		c.Assert(err, IsNil)
-		c.Check(m, DeepEquals, map[string]string{
-			"snapd_recovery_system": s.mockedSystemSeeds[0].label,
-			"snapd_recovery_mode":   mode,
-		})
-		c.Check(s.restartRequests, DeepEquals, []state.RestartType{state.RestartSystemNow})
-		s.restartRequests = nil
-		s.bootloader.BootVars = map[string]string{}
-	}
+	s.testRequestModeWithRestart(c, []string{"install", "recover"}, s.mockedSystemSeeds[0].label)
 }
 
 func (s *deviceMgrSystemsSuite) TestRequestModeErrInBoot(c *C) {
@@ -441,6 +538,7 @@ func (s *deviceMgrSystemsSuite) TestRequestModeErrInBoot(c *C) {
 	err := s.mgr.RequestSystemAction("20191119", devicestate.SystemAction{Mode: "install"})
 	c.Assert(err, ErrorMatches, `cannot set device to boot into system "20191119" in mode "install": no can do`)
 	c.Check(s.restartRequests, HasLen, 0)
+	c.Check(s.logbuf.String(), Equals, "")
 }
 
 func (s *deviceMgrSystemsSuite) TestRequestModeNotFound(c *C) {
@@ -448,12 +546,14 @@ func (s *deviceMgrSystemsSuite) TestRequestModeNotFound(c *C) {
 	c.Assert(err, NotNil)
 	c.Assert(os.IsNotExist(err), Equals, true)
 	c.Check(s.restartRequests, HasLen, 0)
+	c.Check(s.logbuf.String(), Equals, "")
 }
 
 func (s *deviceMgrSystemsSuite) TestRequestModeBadMode(c *C) {
 	err := s.mgr.RequestSystemAction("20191119", devicestate.SystemAction{Mode: "unknown-mode"})
 	c.Assert(err, Equals, devicestate.ErrUnsupportedAction)
 	c.Check(s.restartRequests, HasLen, 0)
+	c.Check(s.logbuf.String(), Equals, "")
 }
 
 func (s *deviceMgrSystemsSuite) TestRequestModeBroken(c *C) {
@@ -464,6 +564,7 @@ func (s *deviceMgrSystemsSuite) TestRequestModeBroken(c *C) {
 	err = s.mgr.RequestSystemAction("20191119", devicestate.SystemAction{Mode: "install"})
 	c.Assert(err, ErrorMatches, "cannot load seed system: cannot load assertions: .*")
 	c.Check(s.restartRequests, HasLen, 0)
+	c.Check(s.logbuf.String(), Equals, "")
 }
 
 func (s *deviceMgrSystemsSuite) TestRequestModeNonUC20(c *C) {
@@ -471,6 +572,13 @@ func (s *deviceMgrSystemsSuite) TestRequestModeNonUC20(c *C) {
 	err := s.mgr.RequestSystemAction("20191119", devicestate.SystemAction{Mode: "install"})
 	c.Assert(err, ErrorMatches, `cannot set device to boot into system "20191119" in mode "install": system mode is unsupported`)
 	c.Check(s.restartRequests, HasLen, 0)
+	c.Check(s.logbuf.String(), Equals, "")
+}
+
+func (s *deviceMgrSystemsSuite) TestRequestActionNoLabel(c *C) {
+	err := s.mgr.RequestSystemAction("", devicestate.SystemAction{Mode: "install"})
+	c.Assert(err, ErrorMatches, "internal error: system label is unset")
+	c.Check(s.logbuf.String(), Equals, "")
 }
 
 func (s *deviceMgrSystemsSuite) TestRequestModeForNonCurrent(c *C) {
@@ -491,4 +599,231 @@ func (s *deviceMgrSystemsSuite) TestRequestModeForNonCurrent(c *C) {
 	err = s.mgr.RequestSystemAction(s.mockedSystemSeeds[1].label, devicestate.SystemAction{Mode: "recover"})
 	c.Assert(err, Equals, devicestate.ErrUnsupportedAction)
 	c.Check(s.restartRequests, HasLen, 0)
+	c.Check(s.logbuf.String(), Equals, "")
+}
+
+func (s *deviceMgrSystemsSuite) TestRequestInstallForOther(c *C) {
+	devicestate.SetSystemMode(s.mgr, "run")
+	// non run modes use modeenv
+	modeenv := boot.Modeenv{
+		Mode: "run",
+	}
+	err := modeenv.WriteTo("")
+	c.Assert(err, IsNil)
+
+	s.state.Lock()
+	s.state.Set("seeded-systems", []devicestate.SeededSystem{
+		{
+			System:  s.mockedSystemSeeds[0].label,
+			Model:   s.mockedSystemSeeds[0].model.Model(),
+			BrandID: s.mockedSystemSeeds[0].brand.AccountID(),
+		},
+	})
+	s.state.Unlock()
+	// reinstall from different system seed is ok
+	s.testRequestModeWithRestart(c, []string{"install"}, s.mockedSystemSeeds[1].label)
+}
+
+func (s *deviceMgrSystemsSuite) TestRequestAction1618(c *C) {
+	s.setPCModelInState(c)
+	// system mode is unset in 16/18
+	devicestate.SetSystemMode(s.mgr, "")
+	// no modeenv either
+	err := os.Remove(dirs.SnapModeenvFileUnder(dirs.GlobalRootDir))
+	c.Assert(err, IsNil)
+
+	s.state.Lock()
+	s.state.Set("seeded-systems", nil)
+	s.state.Set("seeded", nil)
+	s.state.Unlock()
+	// a label exists
+	err = s.mgr.RequestSystemAction(s.mockedSystemSeeds[0].label, devicestate.SystemAction{Mode: "install"})
+	c.Assert(err, ErrorMatches, "cannot set device to boot .*: system mode is unsupported")
+
+	s.state.Lock()
+	s.state.Set("seeded", true)
+	s.state.Unlock()
+
+	// even with system mode explicitly set, the action is not executed
+	devicestate.SetSystemMode(s.mgr, "run")
+
+	err = s.mgr.RequestSystemAction(s.mockedSystemSeeds[0].label, devicestate.SystemAction{Mode: "install"})
+	c.Assert(err, ErrorMatches, "cannot set device to boot .*: system mode is unsupported")
+
+	devicestate.SetSystemMode(s.mgr, "")
+	// also no UC20 style system seeds
+	for _, m := range s.mockedSystemSeeds {
+		os.RemoveAll(filepath.Join(dirs.SnapSeedDir, "systems", m.label))
+	}
+
+	err = s.mgr.RequestSystemAction(s.mockedSystemSeeds[0].label, devicestate.SystemAction{Mode: "install"})
+	c.Assert(err, ErrorMatches, ".*/seed/systems/20191119: no such file or directory")
+	c.Check(s.logbuf.String(), Equals, "")
+}
+
+func (s *deviceMgrSystemsSuite) TestRebootNoLabelNoModeHappy(c *C) {
+	err := s.mgr.Reboot("", "")
+	c.Assert(err, IsNil)
+
+	m, err := s.bootloader.GetBootVars("snapd_recovery_mode", "snapd_recovery_system")
+	c.Assert(err, IsNil)
+	// requested restart
+	c.Check(s.restartRequests, DeepEquals, []state.RestartType{state.RestartSystemNow})
+	// but no bootloader changes
+	c.Check(m, DeepEquals, map[string]string{
+		"snapd_recovery_system": "",
+		"snapd_recovery_mode":   "",
+	})
+	c.Check(s.logbuf.String(), Matches, `.*: rebooting system\n`)
+}
+
+func (s *deviceMgrSystemsSuite) TestRebootLabelAndModeHappy(c *C) {
+	s.state.Lock()
+	s.state.Set("seeded-systems", []devicestate.SeededSystem{
+		{
+			System:  s.mockedSystemSeeds[0].label,
+			Model:   s.mockedSystemSeeds[0].model.Model(),
+			BrandID: s.mockedSystemSeeds[0].brand.AccountID(),
+		},
+	})
+	s.state.Unlock()
+
+	err := s.mgr.Reboot("20191119", "install")
+	c.Assert(err, IsNil)
+
+	m, err := s.bootloader.GetBootVars("snapd_recovery_mode", "snapd_recovery_system")
+	c.Assert(err, IsNil)
+	c.Check(m, DeepEquals, map[string]string{
+		"snapd_recovery_system": "20191119",
+		"snapd_recovery_mode":   "install",
+	})
+	c.Check(s.restartRequests, DeepEquals, []state.RestartType{state.RestartSystemNow})
+	c.Check(s.logbuf.String(), Matches, `.*: rebooting into system "20191119" in "install" mode\n`)
+}
+
+func (s *deviceMgrSystemsSuite) TestRebootModeOnlyHappy(c *C) {
+	s.state.Lock()
+	s.state.Set("seeded-systems", []devicestate.SeededSystem{
+		{
+			System:  s.mockedSystemSeeds[0].label,
+			Model:   s.mockedSystemSeeds[0].model.Model(),
+			BrandID: s.mockedSystemSeeds[0].brand.AccountID(),
+		},
+	})
+	s.state.Unlock()
+
+	for _, mode := range []string{"recover", "install"} {
+		s.restartRequests = nil
+		s.bootloader.BootVars = make(map[string]string)
+		s.logbuf.Reset()
+
+		err := s.mgr.Reboot("", mode)
+		c.Assert(err, IsNil)
+
+		m, err := s.bootloader.GetBootVars("snapd_recovery_mode", "snapd_recovery_system")
+		c.Assert(err, IsNil)
+		c.Check(m, DeepEquals, map[string]string{
+			"snapd_recovery_system": s.mockedSystemSeeds[0].label,
+			"snapd_recovery_mode":   mode,
+		})
+		c.Check(s.restartRequests, DeepEquals, []state.RestartType{state.RestartSystemNow})
+		c.Check(s.logbuf.String(), Matches, fmt.Sprintf(`.*: rebooting into system "20191119" in "%s" mode\n`, mode))
+	}
+}
+
+func (s *deviceMgrSystemsSuite) TestRebootFromRecoverToRun(c *C) {
+	modeenv := boot.Modeenv{
+		Mode:           "recover",
+		RecoverySystem: s.mockedSystemSeeds[0].label,
+	}
+	err := modeenv.WriteTo("")
+	c.Assert(err, IsNil)
+
+	devicestate.SetSystemMode(s.mgr, "recover")
+	err = s.bootloader.SetBootVars(map[string]string{
+		"snapd_recovery_mode":   "recover",
+		"snapd_recovery_system": s.mockedSystemSeeds[0].label,
+	})
+	c.Assert(err, IsNil)
+
+	s.state.Lock()
+	s.state.Set("seeded-systems", []devicestate.SeededSystem{
+		{
+			System:  s.mockedSystemSeeds[0].label,
+			Model:   s.mockedSystemSeeds[0].model.Model(),
+			BrandID: s.mockedSystemSeeds[0].brand.AccountID(),
+		},
+	})
+	s.state.Unlock()
+
+	err = s.mgr.Reboot("", "run")
+	c.Assert(err, IsNil)
+
+	m, err := s.bootloader.GetBootVars("snapd_recovery_mode", "snapd_recovery_system")
+	c.Assert(err, IsNil)
+	c.Check(m, DeepEquals, map[string]string{
+		"snapd_recovery_mode":   "run",
+		"snapd_recovery_system": s.mockedSystemSeeds[0].label,
+	})
+	c.Check(s.restartRequests, DeepEquals, []state.RestartType{state.RestartSystemNow})
+	c.Check(s.logbuf.String(), Matches, fmt.Sprintf(`.*: rebooting into system "%s" in "run" mode\n`, s.mockedSystemSeeds[0].label))
+}
+
+func (s *deviceMgrSystemsSuite) TestRebootAlreadyInRunMode(c *C) {
+	devicestate.SetSystemMode(s.mgr, "run")
+
+	s.state.Lock()
+	s.state.Set("seeded-systems", []devicestate.SeededSystem{
+		{
+			System:  s.mockedSystemSeeds[0].label,
+			Model:   s.mockedSystemSeeds[0].model.Model(),
+			BrandID: s.mockedSystemSeeds[0].brand.AccountID(),
+		},
+	})
+	s.state.Unlock()
+
+	// we are already in "run" mode so this should just reboot
+	err := s.mgr.Reboot("", "run")
+	c.Assert(err, IsNil)
+
+	m, err := s.bootloader.GetBootVars("snapd_recovery_mode", "snapd_recovery_system")
+	c.Assert(err, IsNil)
+	c.Check(m, DeepEquals, map[string]string{
+		"snapd_recovery_mode":   "",
+		"snapd_recovery_system": "",
+	})
+	c.Check(s.restartRequests, DeepEquals, []state.RestartType{state.RestartSystemNow})
+	c.Check(s.logbuf.String(), Matches, `.*: rebooting system\n`)
+}
+
+func (s *deviceMgrSystemsSuite) TestRebootUnhappy(c *C) {
+	s.state.Lock()
+	s.state.Set("seeded-systems", []devicestate.SeededSystem{
+		{
+			System:  s.mockedSystemSeeds[0].label,
+			Model:   s.mockedSystemSeeds[0].model.Model(),
+			BrandID: s.mockedSystemSeeds[0].brand.AccountID(),
+		},
+	})
+	s.state.Unlock()
+
+	errUnsupportedActionStr := devicestate.ErrUnsupportedAction.Error()
+	for _, tc := range []struct {
+		systemLabel, mode string
+		expectedErr       string
+	}{
+		{"", "unknown-mode", errUnsupportedActionStr},
+		{"unknown-system", "run", `stat /.*: no such file or directory`},
+		{"unknown-system", "unknown-mode", `stat /.*: no such file or directory`},
+	} {
+		s.restartRequests = nil
+		s.bootloader.BootVars = make(map[string]string)
+
+		err := s.mgr.Reboot(tc.systemLabel, tc.mode)
+		c.Assert(err, ErrorMatches, tc.expectedErr)
+
+		c.Check(s.restartRequests, HasLen, 0)
+		c.Check(s.logbuf.String(), Equals, "")
+	}
+	c.Check(s.logbuf.String(), Equals, "")
 }
